"""
Dataframe numpy serialization
"""
import typing as t

import numpy.typing as npt
import pyarrow  # type: ignore


def numpy_deserialize(*, value: pyarrow.Tensor) -> npt.NDArray[t.Any]:
    """
    Deserialize a numpy array

    Parameters
    ----------
    value : pyarrow.Tensor
        value to deserialize

    Returns
    -------
    npt.NDArray[t.Any]
        deserialized numpy array
    """
    return t.cast(npt.NDArray[t.Any], value.to_numpy())


def numpy_serialize(*, value: npt.NDArray[t.Any]) -> pyarrow.Tensor:
    """
    Serialize a numpy array

    Parameters
    ----------
    value : npt.NDArray[t.Any]
        value to serialize

    Returns
    -------
    pyarrow.Tensor
        serialized numpy array
    """
    return pyarrow.Tensor.from_numpy(value)



"""
Dataframe arrow serialization
"""
import pandas as pd  # type: ignore
import pyarrow  # type: ignore


def dataframe_deserialize(*, value: pyarrow.Table) -> pd.DataFrame:
    """
    Deserialize a dataframe

    Parameters
    ----------
    value : pyarrow.Table
        value to deserialize

    Returns
    -------
    pd.DataFrame
        deserialized dataframe
    """
    return value.to_pandas()


def dataframe_serialize(*, value: pd.DataFrame) -> pyarrow.Table:
    """
    Serialize a dataframe

    Parameters
    ----------
    value : pd.DataFrame
        value to serialize

    Returns
    -------
    pyarrow.Table
        serialized dataframe
    """
    return pyarrow.Table.from_pandas(value, preserve_index=True)



"""
REST serialization
"""
import typing as t
from functools import singledispatch

import numpy as np
import numpy.typing as npt
import pandas as pd  # type: ignore
import pydantic

from taktile_client.exceptions import SerializationError

from .dataframe import (
    dataframe_deserialize,
    dataframe_serialize,
    dataframe_to_example,
)
from .numpy import numpy_deserialize, numpy_serialize, numpy_to_example
from .pydantic_model import pydantic_model_serialize
from .series import series_deserialize, series_serialize, series_to_example
from .utils import JSONStructure


@singledispatch
def serialize(
    value: t.Union[
        npt.NDArray[t.Any], pd.Series, pd.DataFrame, pydantic.BaseModel
    ]
) -> str:
    """
    Serialize object for REST

    Parameters
    ----------
    value : t.Union[
        npt.NDArray[t.Any], pd.Series, pd.DataFrame, pydantic.BaseModel
    ]
        value to serialize

    Returns
    -------
    str
        serialized object
    """
    raise SerializationError(f"Can't serialize value of type {type(value)}")


@serialize.register
def _serialize_numpy(value: np.ndarray) -> str:  # type: ignore
    return numpy_serialize(value=value)


@serialize.register
def _serialize_series(value: pd.Series) -> str:
    return series_serialize(value=value)


@serialize.register
def _serialize_dataframe(value: pd.DataFrame) -> str:
    return dataframe_serialize(value=value)


@serialize.register
def _serialize_pydantic_model(value: pydantic.BaseModel) -> str:
    return pydantic_model_serialize(value=value)


@singledispatch
def deserialize(
    sample: t.Union[npt.NDArray[t.Any], pd.Series, pd.DataFrame],
    *,
    value: JSONStructure,
) -> t.Union[npt.NDArray[t.Any], pd.Series, pd.DataFrame]:
    """
    Deserialize object

    Parameters
    ----------
    sample : InputType
        sample object
    value : JSONStructure
        value to deserialize

    Returns
    -------
    InputType
        deserialized object
    """
    raise SerializationError(f"Can't deserialize value of type {type(sample)}")


@deserialize.register
def _deserialize_numpy(
    sample: np.ndarray, *, value: JSONStructure  # type: ignore
) -> npt.NDArray[t.Any]:
    return numpy_deserialize(value=value, sample=sample)


@deserialize.register
def _deserialize_series(
    sample: pd.Series, *, value: JSONStructure
) -> pd.Series:
    return series_deserialize(value=value, sample=sample)


@deserialize.register
def _deserialize_dataframe(
    sample: pd.DataFrame, *, value: JSONStructure
) -> pd.DataFrame:
    return dataframe_deserialize(value=value, sample=sample)


@singledispatch
def to_example(
    value: t.Union[npt.NDArray[t.Any], pd.Series, pd.DataFrame]
) -> str:
    """
    Produce serialized example of object

    Parameters
    ----------
    value : t.Union[npt.NDArray[t.Any], pd.Series, pd.DataFrame]
        value to produce an example of

    Returns
    -------
    str
        serialized example
    """
    raise SerializationError(
        f"Can't create example value of type {type(value)}"
    )


@to_example.register
def _to_example_numpy(value: np.ndarray) -> str:  # type: ignore
    return numpy_to_example(value=value)


@to_example.register
def _to_example_series(value: pd.Series) -> str:
    return series_to_example(value=value)


@to_example.register
def _to_example_dataframe(value: pd.DataFrame) -> str:
    return dataframe_to_example(value=value)



"""
Taktile Client
"""
__version__ = "1.0.0"

from .config import arrow_available

if arrow_available:
    from taktile_client.arrow import ArrowClient

# pylint: disable-next=wrong-import-position, wrong-import-order
from taktile_client.rest import RestClient



"""
Series arrow serialization
"""
import pandas as pd  # type: ignore
import pyarrow  # type: ignore


def series_deserialize(*, value: pyarrow.Array) -> pd.Series:
    """
    Deserialize a series

    Parameters
    ----------
    value : pyarrow.Array
        value to deserialize

    Returns
    -------
    pd.Series
        deserialized series
    """
    return value.to_pandas()


def series_serialize(*, value: pd.Series) -> pyarrow.Array:
    """
    Serialize a Series

    Parameters
    ----------
    value : pd.Series
        value to serialize

    Returns
    -------
    pyarrow.Array
        serialized series
    """
    return pyarrow.Array.from_pandas(value)



"""
Arrow serialization
"""
import typing as t
from functools import singledispatch

import numpy as np
import numpy.typing as npt
import pandas as pd  # type: ignore
import pyarrow  # type: ignore

from taktile_client.exceptions import SerializationError

from .dataframe import dataframe_deserialize, dataframe_serialize
from .numpy import numpy_deserialize, numpy_serialize
from .series import series_deserialize, series_serialize

SerialType = t.Union[pyarrow.Table, pyarrow.Array, pyarrow.Tensor]
UserType = t.Union[npt.NDArray[t.Any], pd.Series, pd.DataFrame]


@singledispatch
def serialize(value: UserType) -> SerialType:
    """
    Serialize object for REST

    Parameters
    ----------
    value : t.Union[npt.NDArray[t.Any], pd.Series, pd.DataFrame]
        value to serialize

    Returns
    -------
    str
        serialized object
    """
    raise SerializationError(f"Can't serialize value of type {type(value)}")


@serialize.register
def _serialize_numpy(value: np.ndarray) -> pyarrow.Tensor:  # type: ignore
    return numpy_serialize(value=value)


@serialize.register
def _serialize_series(value: pd.Series) -> pyarrow.Array:
    return series_serialize(value=value)


@serialize.register
def _serialize_dataframe(value: pd.DataFrame) -> pyarrow.Table:
    return dataframe_serialize(value=value)


@singledispatch
def deserialize(
    value: SerialType,
) -> UserType:
    """
    Deserialize value

    Parameters
    ----------
    value : SerialType
        value to deserialize

    Returns
    -------
    UserType
        deserialized object
    """
    raise SerializationError(f"Can't deserialize value of type {type(value)}")


@deserialize.register
def _deserialize_numpy(value: pyarrow.Tensor) -> npt.NDArray[t.Any]:
    return numpy_deserialize(value=value)


@deserialize.register
def _deserialize_series(value: pyarrow.Array) -> pd.Series:
    return series_deserialize(value=value)


@deserialize.register
def _deserialize_dataframe(value: pyarrow.Table) -> pd.DataFrame:
    return dataframe_deserialize(value=value)
